/*==
Pov-Ray   : 3.8
Scene File: FABRIK.inc
Author    : Ingo Janssen
Date      : 01-12-2018
Rev Date  : 10-12-2018
Version   : 1a4 unify the solvers
-- 
Inverse Kinematic Framework include file for POV-ray based on:
Forward And Backward Reaching Inverse Kinematics (FABRIK) by Andreas Aristidou:
  http://www.andreasaristidou.com/FABRIK.html
  http://www.andreasaristidou.com/publications/papers/FABRIK.pdf
  http://www.andreasaristidou.com/publications/papers/Extending_FABRIK_with_Model_C%CE%BFnstraints.pdf
Excellent code inspiration & guide, by unknown author:
  https://developer.roblox.com/articles/Inverse-Kinematics-for-Animation#FABRIK

As ray tracing isn't exactly real time, several of the optimalisations as mentioned
in the second paper are left out (so far)
--
macros starting with an underscore should not be called 'from the outside'.
only chains mentioned in the process dictionary will be initialised
absence of the process dictionary will result in an error

The FABRIK macro needs a rather complex datastructure as input, to be explained
later. Create an 'instance' of the Fabrik 'object': #declare S = FABRIK(F); 
where F is the datastructure. Fabrik 'returns' an updated/initialised version
of the input structure. 

To run the IK solver use SetTarget(S, T) where S is the afore mentioned
instance of Fabrik and T a datastructure holding the target(s)
*/
/*
A FABRIK 'object' consists of a data structure containing:

  'Index' (dict): 
    'Chains' (dict): Holds the names of all the chains in the FABRIK object.
    'SubBases'(dict): Holds the names of all the subbases, the points where chains join.
  
  'Proc' (array): 
    Holds the names of chains, sub bases and closed loops in the order they 
    should be processed in. 
  
  'Chains' (dict): 
    Chains are named ordered collections of links & joints. Thinking in bones,
    a bone in IK is a joint and a link with a length. A bone has only one
    joint at its start point.  J----
  
    'Joint' (array): 
      holding all the joints (<vector>) and the initial target or EndPoint.
      J----J--J----T
    
    '_Lengths' (array):
      On initialisation of a Joint array an array of links _Lengths is created
      4----2--4----
    
    'Type' (string): 
      Types of chains and the indicator for a SubBase
      if it is part of this chain. The types are always named 'from to',
      so BaseEnd array goes from base point [0] to end point [max] and 
      holds max+1 joints.
      Available Types:
        BaseEnd        (Just a Trunk. Can have a Leaf)
        BaseSubbase    (Trunk)
        SubbaseSubbase (Branch)
        SubbaseEnd     (Twig)
        ClosedLoop
    
    'Constraints' (array, optional): 
      It holds references of the movement constraint objects per Joint, or false
      if uncontrained.
      Not implemented yet.
  
  'Subbase' (dict):
    'Index' (arr): 
      lists the chains that are connected to the subbase in the order
      that they are processed.
    'Centroid' (vec): 
      holds one points that is exchanged between the connected chains,
      initially it is <0,0,0>
    'Type' (string): "SubBase"

*/

#version 3.8;

#include "arrays.inc"

#macro FABRIK(self)
  //==
  // Framework for dealing with IK chains, for different 
  // constellations of chains of joints.
  //
  // FABRIK(self) parameters:
  // self (dict) : datastructure containing chains with joints and procecssing
  // instructions for the IK solvers
  //
  #ifndef (self.Proc)
    #error "\nNo Process 'Proc' dictionary defined\n"
  #end
  
  // add FABRIK global defaults, if not defined.
  #ifndef (self.Default)
    #local self.Default = dictionary {
      ["Tolerance"] : 0.1,
      ["MaxIter"] : 100
    };
  #else
    #ifndef(self.Default.Tolerance)
      #local self.Default.Tolerance = 0.1;
    #end
    #ifndef(self.Default.MaxIter)
      #local self.Default.MaxIter = 100;
    #end
  #end
  #local self.ProcessType = dictionary{
    // Chain types
    ["BaseEnd"] : 0,      
    ["BaseSubbase"] : 1,  
    ["SubbaseSubbase"] : 2,
    ["SubbaseEnd"] : 3,   
    ["ClosedLoop"] : 4, //not implemented yet
    // Processing
    ["SubBase"] : 5,
    ["Connector"]: 6    //not implemented yet
  }
  
  #macro _Init(self)
    //==
    // Initialise the Fabrik 'object'. 
    //
    // _Init(self) parameters:
    // self (dict) : Fabrik object
    //
    #macro _LinkLengths(self, Chain) 
      //==
      // Calculates the lengths between the joints, adds them to the current
      // chain Adds up all the lengths and adds the result to the current chain
      //
      // Lives within _Init for now as long as there are no prismatic 
      // (extending) joints implemented. Changing distances between joints
      // would require re-initialisation.
      //
      // _LinkLengths(self, Chain)
      // self  (dict)  : Fabrik object
      // Chain (string): name of the current chain
      //
      #local self[Chain]._Lengths = array[self[Chain]._N-1];
      #local self[Chain]._Length = 0;
      #local i = 0;
      #while (i < (self[Chain]._N-1))
        #local self[Chain]._Lengths[i] = 
          vlength(self[Chain].Joint[i]-self[Chain].Joint[i+1]);
        #local self[Chain]._Length = 
          self[Chain]._Length + self[Chain]._Lengths[i];
        #local i = i + 1; 
      #end
    #end
    
    #macro _InitChain(self, Chain)
      //==
      // Initialise a Chain, add an array with lengths 
      // between joints and add the total length. For quicker access add origin
      // of the chain _O number of joints in the chain _N and _M=_N-1
      // 
      // _InitChain(self, Chain) parameters:
      // self  (dict)  : Fabrik object
      // Chain (string): name of the current chain
      //
      #debug concat("init chain : ", Chain, "\n")
      #local self[Chain]._O = self[Chain].Joint[0];
      #local self[Chain]._N = 
        dimension_size(self[Chain].Joint, 1);
      #local self[Chain]._M = self[Chain]._N-1;          //max index of array
      // pre set effector and Target to EndPoint as every Chain needs a target
      // to be solved even if the 'pose' doesn't change.
      #local self[Chain].Effector = self[Chain]._M; 
      #local self[Chain].Target = self[Chain].Joint[self[Chain].Effector];
      _LinkLengths(self, Chain)
    #end
    
    #debug "\nInitialise :\n"
    #for (i,0,dimension_size(self.Index.Chains,1)-1,1)
      #local Chain = self.Index.Chains[i];
      _InitChain(self, Chain)
    #end
    #debug "\n"
  #end
  _Init(self)


  #macro _Reach(self, Chain)
    //==
    // Checks whether a target is within reach of the current chain.
    //
    // _Reach(self, Chain) parameters:
    // self  (dict)  : Fabrik object
    // Chain (string): name of the current chain
    //
    #debug "  Reach : "
    
    #if (vlength(self[Chain]._O - self[Chain].Target) > self[Chain]._Length)
      #local self[Chain]._Reachable = false;
      #debug "False\n"
    #else
      #local self[Chain]._Reachable = true;
      #debug "True\n"
    #end
  #end

  #macro _Backward(self, Chain)
    //==
    // Steps backwards through the joints in the current chain and adjusts
    // their positions.
    //
    // _Backwards(self, Chain) parameters:  
    // self  (dict)  : Fabrik object
    // Chain (string): name of the current chain
    //
    #debug concat("Backward : ",Chain,"\n")
    #local self[Chain].Joint[self[Chain].Effector] = self[Chain].Target;
    #for (i, self[Chain]._M-1, 0, -1)
      #local l = self[Chain]._Lengths[i]/
        vlength(self[Chain].Joint[i+1]-self[Chain].Joint[i]);
      #local Pos =  (1-l) * self[Chain].Joint[i+1] + l * self[Chain].Joint[i];
      //if constraint
      //check constraint
      //if not within constraint adjust Pos
      //2B done
      #local self[Chain].Joint[i] = Pos;
      #debug concat("i : ",str(i,0,0)," pos : ", vstr(3,Pos,",",0,3),"\n")
    #end
  #end

  #macro _Forward(self, Chain)
    //==
    // Steps forwards through the joints in the current chain and adjusts their
    // positions.
    //
    // _Forward(self, Chain) parameters:  
    // self  (dict)  : Fabrik object
    // Chain (string): name of the current chain
    //
    #debug concat("Forward : ",Chain,"\n")
    // set first point back to its original position
    #local self[Chain].Joint[0] = self[Chain]._O; 
    #for (i, 0, self[Chain]._M-1, 1)
      #local l = self[Chain]._Lengths[i]/
        vlength(self[Chain].Joint[i+1]-self[Chain].Joint[i]);
      #local Pos =  (1-l) * self[Chain].Joint[i] + l * self[Chain].Joint[i+1];
      //if constraint
      //check constraint
      //if not within constraint adjust Pos
      //2B done
      #local self[Chain].Joint[i+1] = Pos;
      #debug concat("i : ",str(i,0,0)," pos : ", vstr(3,Pos,",",0,3),"\n")
    #end
  #end

  #macro _Stretch(self, Chain)
    //==
    // Stretches a chain to full lengths in the direction of the target. Used
    // when the target is not reachable.
    //
    // _Stretch(self, Chain) parameters:
    // self  (dict)  : Fabrik object
    // Chain (string): name of the current chain
    //
    #debug "Stretch :\n"
    #for (i, 0, self[Chain]._M-1, 1)
      #local l = self[Chain]._Lengths[i]/
        vlength(self[Chain].Target-self[Chain].Joint[i]);  
      #local self[Chain].Joint[i+1] = (1-l) * self[Chain].Joint[i]
        + l * self[Chain].Target;
      #debug concat(
        "i : ", str(i,0,0), " pos : ", vstr(3,self[Chain].Joint[i+1],",",0,3),
        "\n"
      )
    #end
  #end

  #macro _Solve(self)
    //==
    // Traverses the Proc array and solves each chain step by step until all 
    // chains are solved or MaxIter is hit. It traverses from left to right
    // and right to left always including both ends a,b,c c,b,a a,b,c c,b,a. 
    //
    // _Solve(self) parameters:
    // self (dict): Fabrik data structure
    //
    #debug "Solve  :\n"
    #local ProcDim = dimension_size(self.Proc,1);
    // Create a "point of truth"
    // Set all chains to solved
    #for (i, 0, dimension_size(self.Index.Chains,1)-1, 1)
      #local self[self.Index.Chains[i]].Solved = true;
    #end
    // set all to be processed chains to not solved
    #for (i, 0, ProcDim-1,1)
      #local self[self.Proc[i]].Solved = false;
    #end
    #local Direction = "LeftRight";
    #debug "\nDirection LeftRight\n"
    #local i = 0;
    #local SolveIter = 1;
    #while(SolveIter < self.Default.MaxIter)
      #local Process = self.Proc[i];
      #ifdef (self[Process].Type)
        #local PT = self.ProcessType[self[Process].Type]; 
        #switch(PT)
        
          #case(self.ProcessType["BaseEnd"])
            #local Chain = Process;
            #if(self[Chain].Solved = false)
              #debug "ProcessType : BaseEnd\n"
              #ifndef (self[Chain]._Reachable)
                _Reach(self, Chain)
              #end
              #if (self[Chain]._Reachable = false)
                _Stretch(self, Chain)
                #local self[Chain].Solved = true;
                #debug concat("Chain : ", Chain, " solved\n") 
              #else
                #if(Direction = "LeftRight")
                  _Backward(self, Chain)            
                #elseif(Direction = "RightLeft")
                  _Forward(self, Chain)
                  #local dif = vlength(self[Chain].Joint[self[Chain].Effector] - self[Chain].Target);
                  #if(dif <= self.Default.Tolerance)
                    #local self[Chain].Solved = true;
                    #debug concat("Chain : ", Chain, " solved\n") 
                  #end
                #end
              #end
            #end
          #break
          
          #case(self.ProcessType["BaseSubbase"])
            #local Chain = Process;
            #debug "ProcessType : BaseSubbase\n"
            #if(self[Chain].Solved = false)
              #if(Direction = "LeftRight")
                #local Sub = self[Chain].SubBase[0];
                #local Centroid = self[Sub].Centroid;
                #local self[Chain].Target = Centroid;
                #debug concat("Get centroid from ", Sub," set target to ", Chain," : ", vstr(3,self[Chain].Target,",",0,3) ,"\n")
                _Backward(self, Chain)
              #elseif(Direction = "RightLeft")
                _Forward(self, Chain)
              #end
            #end
          #break

          #case(self.ProcessType["SubbaseSubbase"])   
            #local Chain = Process;
            #debug "ProcessType : SubbaseSubbase\n"
            #if(self[Chain].Solved = false)
              #if(Direction = "LeftRight")
                #local Sub = self[Chain].SubBase[0];
                #local Centroid = self[Sub].Centroid;
                #local self[Chain].Target = Centroid;
                #debug concat("Get centroid from ",Sub," set Target to ", Chain," : ", vstr(3,self[Chain].Target,",",0,3) ,"\n")
                _Backward(self, Chain)
              #elseif(Direction = "RightLeft")
                #local Sub = self[Chain].SubBase[1];
                #local Centroid = self[Sub].Centroid;
                #local self[Chain]._O = Centroid;
                #debug concat("Get centroid from ",Sub," set Origin to ", Chain," : ", vstr(3,self[Chain]._O,",",0,3) ,"\n")
                _Forward(self, Chain)
              #end
            #end
          #break

          #case(self.ProcessType["SubbaseEnd"])   
            #local Chain = Process;
            #debug "ProcessType : SubbaseEnd\n"
            #if(self[Chain].Solved = false)
              #if(Direction = "LeftRight")
                _Backward(self, Chain)
              #elseif(Direction = "RightLeft")
                #local Sub = self[Chain].SubBase[0];
                #local Centroid = self[Sub].Centroid;
                #local self[Chain]._O = Centroid;
                #debug concat("Get centroid from ",Sub," set Origin to ",Chain," : ", vstr(3,self[Chain]._O,",",0,3) ,"\n")
                _Forward(self, Chain)
              #end
            #end
          #break
          
          #case(self.ProcessType["ClosedLoop"])
            #debug "ProcessType : ClosedLoop\n"
            // 2Bdone
          #break
          
          #case(self.ProcessType["SubBase"])
            #local Subbase = Process;
            #debug "ProcessType : SubBase\n"
            #if(self[Subbase].Solved = false)
              #local L=dimension_size(self[Subbase].Index,1)-2;
              #if(Direction = "LeftRight")
                // calc centroid
                #local Centroid = <0,0,0>;
                #for (n,0,L,1)
                  #local Centroid = Centroid + self[self[Subbase].Index[n]].Joint[0]; 
                  #debug concat("SubBase ", Subbase," get point from ",self[Subbase].Index[n] ," add ", vstr(3,Centroid,",",0,3) ,"\n")
                #end
                #undef n
                #local self[Subbase].Centroid = Centroid/(L+1);
                #debug concat("SubBase ", Subbase,"  self.Centroid : ", vstr(3,self[Subbase].Centroid,",",0,3) ,"\n")
              #elseif(Direction = "RightLeft")
                // set centroid to vec of last point in first chains
                #local Chain = self[Subbase].Index[L+1]; 
                #local self[Subbase].Centroid = self[Chain].Joint[self[Chain]._M];
                #debug concat("SubBase  ", Subbase," get point from ", Chain ,"set Centroid : ", vstr(3,self[Subbase].Centroid,",",0,3) ,"\n")
              #end
            #end
          #break
          
          #case(self.ProcessType["Connector"])
            #debug "ProcessType : Connector\n"
          #break
          
        #end
      #end
      // traverse the Proc array from left to right and right to left
      // always including both ends a,b,c c,b,a a,b,c c,b,a
      #if(Direction = "LeftRight")
        #local i = i+1;
      #elseif(Direction = "RightLeft")
        #local i = i-1;
      #end
      #if (i = ProcDim)
        #local Direction = "RightLeft";
        #debug "\nDirection change to RightLeft\n"
        #local i = i-1;
      #elseif(i = -1)
        #local Direction = "LeftRight";
        #debug concat("Iteration : ", str(SolveIter,0,0), "\n\n")
        // After each iteration,
        // Check solved chains, first whether all targets are reached and then wether
        // all chains are solved. All interlinked chains have to be checked and set to
        // Solved=true at once otherwise some parts can't be resolved anymore.
        // Independent chains 'BaseEnd' can set them selves solved and once solved
        // they are not visited by the solver anymore.
        #local TargetsReached = 0;
        #local TR = 0;
        #for (n, 0, dimension_size(self.Index.Chains,1)-1, 1)
          #if (self[self.Index.Chains[n]].Solved = false)
            #local TR = TR+1;
            #local dif = 
                vlength(self[Chain].Joint[self[Chain].Effector] - self[Chain].Target);
            #if (dif <= self.Default.Tolerance)
              #local TargetsReached = TargetsReached+1;
            #end
          #end
        #end
        #if (TargetsReached = TR)
          #for (n, 0, dimension_size(self.Index.Chains,1)-1, 1)
            #if (self[self.Index.Chains[n]].Solved = false)
              #local self[self.Index.Chains[n]].Solved = true;
            #end
          #end
        #end
        //Check all chains for Solved state
        #local SolvedChain = 0;
        #for (n, 0, dimension_size(self.Index.Chains,1)-1, 1)
          #if (self[self.Index.Chains[n]].Solved = true)
            #local SolvedChain = SolvedChain+1;
          #end
        #end
        #undef n
        #if (SolvedChain = dimension_size(self.Index.Chains,1))
          #break
        #end
        #debug "\nDirection change to LeftRight\n"
        #local i = i+1;
        #local SolveIter = SolveIter+1;
      #end
    #end
  #end
  
  #macro StickMan(self)
    //==
    // Puts a stick man in the scene, nodes are rendered as spheres, links as
    // cylinders.
    //
    // 
    #for (i,0,dimension_size(self.Index.Chains,1)-1,1)
      #ifdef (self[self.Index.Chains[i]].Joint)
        #local Chain = self.Index.Chains[i];
        #for (n,0,self[Chain]._M-1,1)    
          sphere{self[Chain].Joint[n], 0.15 pigment{rgb z}}
          cylinder{self[Chain].Joint[n],self[Chain].Joint[n+1],0.1}
        #end
        sphere{self[Chain].Target, 0.151 pigment{rgb x}} //end points
      #end 
    #end
  #end

  #macro PinMan(self)
    //==
    // Puts a pin man in the scene, links are rendered as two cones.
    //
    // 
    #for (i,0,dimension_size(self.Index.Chains,1)-1,1)
      #ifdef (self[self.Index.Chains[i]].Joint)
        #local Chain = self.Index.Chains[i];
        #for (n,0,self[Chain]._M-1,1)    
          #local Mid = (self[Chain].Joint[n]+self[Chain].Joint[n+1])/2;
          cone{self[Chain].Joint[n],0,Mid,0.1}
          cone{Mid,0.1,self[Chain].Joint[n+1],0}
          sphere{self[Chain].Joint[n], 0.05 pigment{rgb z}} //end points
        #end
        sphere{self[Chain].Target, 0.051 pigment{rgb x}} //end points
      #end 
    #end
  #end

  #macro SplineMan(self)
    //==
    // Puts a natural_spline man in the scene, nodes are rendered as spheres, links as
    // a spline of spheres.
    //
    //   
    #for (i,0,dimension_size(self.Index.Chains,1)-1,1)
      #ifdef (self[self.Index.Chains[i]].Joint)
        #local Chain = self.Index.Chains[i];
        #local Spl = spline{
          natural_spline
          #for (n,0,self[Chain]._M,1)    
            n, self[Chain].Joint[n]
          #end
        }
        #for (n,0,self[Chain]._M,0.01)
          sphere{Spl(n),0.1}
        #end
        #for (n,0,self[Chain]._M-1,1)
          sphere{Spl(n),0.15 pigment{rgb z}}
        #end
        sphere{self[Chain].Target,0.151 pigment{rgb x}}
      #end 
    #end
  #end

  #macro SetTarget(self, T)
    //==
    // Sets the target for the IK chain(s) and starts the solver. Targets
    // can be EndPoints, Origins or Leaves (2B done)
    //
    // SetTarget (self, T) parameters:
    // self (dict): The previously declared Fabrik object
    // T (dict)   : The datastructure holding the target(s) for chain(s)
    //
    #debug "Set Targets:\n"
    #for (i, 0, dimension_size(T.Index,1)-1,1)
      #local Chain = T.Index[i];
      #debug concat("set target : ", Chain)
      #local TargetDim = dimension_size(T[Chain],1);
      #if (TargetDim > 2)
        #error "Target can't have a length of 3 or more!"
      #elseif (TargetDim = 1)
        // Single entry so we'll assume effector is an EndPoint
        #local self[Chain].Target = T[Chain][0];
        #local self[Chain].Effector = self[Chain]._M;
        #debug concat("  set effector EndPoint to Target: ", vstr(3,self[Chain].Target,",",0,3), "\n")        
      #elseif (TargetDim = 2)
          #local Eff = T[Chain][0];
          #local Tar = T[Chain][1];
          #if (Eff = 0)
            #local self[Chain]._O = Tar;
            #if (self[Chain].Type = "BaseEnd")
              #local self[Chain].Effector = Eff;
            #end
            #debug concat("  set Origin to Point           : ", vstr(3,self[Chain].Target,",",0,3), "\n")
          #else
            #local self[Chain].Effector = Eff;
            #local self[Chain].Target = Tar;
            #if (Eff = self[Chain]._M)
              #debug concat("  set effector EndPoint to Target: ", vstr(
                3,self[Chain].Target,",",0,3), "\n"
              )
            #else
              #debug concat("  set effector Leaf to Target    : ", vstr(
                3,self[Chain].Target,",",0,3), "\n"
              )
            #end
          #end
      #else
        #error "Can't be done, sorry"        
      #end
    #end
    #debug "\n"
    _Solve(self)
  #end
  
  // 'return' self, the Fabrik object
  self
#end


/*test scene ----------------------------------------------*/

#if(input_file_name = "fabrik.inc")
//Scenes "I", "Four", "Y", "YY"
#declare Scene = "I";


#global_settings{assumed_gamma 1.0 }
#default{pigment{rgb 1} finish{ ambient 0.2 diffuse 0.9 }} 
light_source{<1000,1000,-1000>, rgb 1}
plane{y,0 pigment{checker rgb 0.05 rgb 0.1}}


//==
// Test scene 'I'. Simplest possible scene, just a stick of whitch the 
// target and origin can be moved.
//
// F is the datastruktje that FABRIK needs. These are the minum required items.
// "Index" (dict): as POV-Ray can't loop over dictionaries (yet) we need to index 
//         the content. The Chains item in the index is required and holds the 
//         names of the chains.
// "Proc" (array): holds the processing instructions, and defines the order in 
//         which they should be processed. In this scene just the chain 'Base1'.
// "Base1" (dict): Base1 is the name of a chain.
//          "Joint" (array): joints
//          "Type" (string": type is required for processing, in this case 'BaseEnd'
//                           just a 'trunk' with nothing attached.
//
// T is the target, a dictionary
//  "Index" (array): holds the targeted Chain names.
//  "Chain Id" (mixed array): 
//         the name of the chain that the target is for and the target point 
//         {effector (=joint index(int)), target(vec)}
//         Normally in an IK you move end points so a shortcut is implemented
//         an array holding just the target vector is interpreted as being the
//         end point of the chain. The effector can be specified to move the 
//         origin of the chain {0,<vec>} or to move a 'Leaf' any node between
//         origin and end point. 
//         Moving Leaves is not implemented yet.
//
// StickMan, PinMan and SplineMan are little helpers to visualise the chains
// comment out the "SetTarget" to see the base pose. The red spheres are the 
// targets 
//
// SetTarget sets the targets for the chains and starts the solver to get to
// the new pose. SetTarget does not reinitialise the FABRIK, so each new set target
// continues with the previos pose.
//
// 
#if (Scene = "I")
  // BaseEnd  test
  camera{location <0,2.5,-10> look_at <0,2.5,0>}
  plane{z,0 pigment{checker rgb 0.05 rgb 0.1}}
  
  #declare F = dictionary{
    ["Index"]: dictionary{["Chains"] : array[1]{"Base1"}},
    ["Proc"]: array[1]{"Base1"},
    ["Base1"] : dictionary {
        ["Joint"] : array[6]{
                      <0,0,0>, <0,1,0>, <0,2,0>, 
                      <0,3,0>, <0,4,0>, <0,5,0>
                    },
        ["Type"]  : "BaseEnd"
    }                                               
  };
  #declare S = FABRIK(F);
  
  #declare T = dictionary {
    ["Index"] : array[1]{"Base1"},
    ["Base1"] : array mixed[2]{5,<4,2,0>}  // = array[1]{<4,2,0>}
  };
  SetTarget(S, T)
  //StickMan(S) 
  //PinMan(S) 
  SplineMan(S) 
#end



//==
// Test scene 'Four'. Four independent, not connected 'BaseEnd' chains.
// Nothing new, just more of the same.
//
#if (Scene = "Four")
  camera{location <3,5,-5> look_at <0,0,0>}
  #declare F = dictionary{
    ["Index"]: dictionary{["Chains"] : array[4]{"Base1", "Base2", "Base3", "Base4"}},
    ["Proc"]: array[4]{"Base1", "Base2", "Base3", "Base4"},
    ["Base1"] : dictionary {
        ["Joint"] : array[6]{
                      <0,0,0>, <0,1,0>, <0,1.5,0>, 
                      <0,3,0>, <0,4,0>, <0,5,0>
                    },
        ["Type"]  : "BaseEnd"
    }                                               
    ["Base2"] : dictionary {                      
        ["Joint"] : array[6]{
                      <0,0,0>, <0,1,0>, <0,1.5,0>, 
                      <0,3,0>, <0,4,0>, <0,5,0>
                    },
        ["Type"]  : "BaseEnd"
    }
    ["Base3"] : dictionary {
        ["Joint"] : array[6]{
                      <0,0,0>, <0,1,0>, <0,1.5,0>, 
                      <0,3,0>, <0,4,0>, <0,5,0>
                    },
        ["Type"]  : "BaseEnd"
    }                                               
    ["Base4"] : dictionary {                      
        ["Joint"] : array[6]{
                      <0,0,0>, <0,1,0>, <0,1.5,0>, 
                      <0,3,0>, <0,4,0>, <0,5,0>
                    },
        ["Type"]  : "BaseEnd"
    }
  };
  #declare S = FABRIK(F);
  
  #declare T = dictionary {
    ["Index"] : array[4]{"Base1", "Base2", "Base3", "Base4"},
    ["Base1"] : array mixed[2]{5,< 4, 0, 0>},
    ["Base2"] : array mixed[2]{5,<-4, 2, 0>},
    ["Base3"] : array mixed[1]{< 0, 0, 4>},
    ["Base4"] : array mixed[1]{< 0, 0,-4>}
  };
  SetTarget(S, T)
  //StickMan(S) 
  SplineMan(S) 
  //PinMan(S)
#end



//==
// Scene 'Y'. a 'tunk' with two branches. The branches can be 'targeted' as
// well as the base of the 'trunk'. The positio of the branch targets 
// infulence the 'trunk'
//
#if (Scene = "Y")
  camera{location <0,4,-10> look_at <0,4,0>}
  plane{z,0 pigment{checker rgb 0.05 rgb 0.1}}
  
  #declare F = dictionary{ 
    ["Index"] : dictionary{
      ["Chains"] : array[3]{"Left1", "Right1", "Base"},
      ["SubBases"]: array[1]{"LRB"}
    },
    ["Proc"] : array[4]{"Left1", "Right1", "LRB" ,"Base"},
    ["Base"] : dictionary {
        ["Joint"]  : array[5]{<0,0,0>, <0,1,0>, <0,2,0>,<0,3,0>, <0,4,0>},
        ["Type"]   : "BaseSubbase",
        ["SubBase"]: array mixed[1]{"LRB"}
    },                                               
    ["Left1"] : dictionary {
        ["Joint"]  : array[4]{<0,4,0>, <-1,5,0>, <-2,6,0>,<-3,7,0>},
        ["Type"]   : "SubbaseEnd",
        ["SubBase"]: array mixed[1]{"LRB"}
    },                                               
    ["Right1"]: dictionary {
        ["Joint"]  : array[4]{<0,4,0>, <1,5,0>, <2,6,0>,<3,7,0>},
        ["Type"]   : "SubbaseEnd",
        ["SubBase"]: array mixed[1]{"LRB"}
    },
    ["LRB"]  : dictionary{
        ["Index"]   : array[3]{"Left1", "Right1", "Base"}
        ["Type"]    : "SubBase",
        ["Centroid"]: <0,0,0>
    }
  };
  #declare S = FABRIK(F);

  #declare T = dictionary {
    ["Index"] : array[2]{"Right1", "Left1",},
    ["Right1"] : array mixed[1]{<3, 4.5, 0>},
    ["Left1"] : array mixed[1]{<-1, 1, 0>},
  };
  SetTarget(S, T)
  //StickMan(S) 
  SplineMan(S) 
  //PinMan(S)
#end



//==
// Scene 'YY'. a 'tunk' with two branches and two twigs. One branche and two twigs
// can be 'targeted' as well as the base of the 'trunk'. The position of the branch targets 
// infulence the 'trunk' etc.
//
#if (Scene = "YY")
  camera{location <0,4.5,-12> look_at <0,4.5,0>}
  plane{z,0 pigment{checker rgb 0.05 rgb 0.1}}
  
  #declare F = dictionary{ 
    ["Index"] : dictionary{
      ["Chains"] : array[5]{"Base", "Left", "Right", "Right1", "Right2"},
      ["SubBases"]: array[2]{"LRB", "RRR"}
    },
    ["Proc"] : array[7]{"Right1", "Right2", "RRR", "Right", "Left",  "LRB" ,"Base"},
    ["Base"] : dictionary{
        ["Joint"]  : array[5]{<0,0,0>, <0,1,0>, <0,2,0>,<0,3,0>, <0,4,0>},
        ["Type"]   : "BaseSubbase",
        ["SubBase"]: array mixed[1]{"LRB"}
    },                                               
    ["Left"] : dictionary{
        ["Joint"]  : array[4]{<0,4,0>, <-1,5,0>, <-2,6,0>,<-3,7,0>},
        ["Type"]   : "SubbaseEnd",
        ["SubBase"]: array mixed[1]{"LRB"}
    },                                               
    ["Right"]: dictionary{
        ["Joint"]  : array[4]{<0,4,0>, <1,5,0>, <2,6,0>,<3,7,0>},
        ["Type"]   : "SubbaseSubbase",
        ["SubBase"]: array mixed[2]{"RRR", "LRB"}
    },
    ["Right1"]: dictionary{
        ["Joint"]  : array[3]{<3,7,0>, <4,8,0>, <5,9,0>},
        ["Type"]   : "SubbaseEnd",
        ["SubBase"]: array mixed[1]{"RRR"}
    },
    ["Right2"]: dictionary{
        ["Joint"]  : array[3]{<3,7,0>, <2,8,0>, <1,9,0>},
        ["Type"]   : "SubbaseEnd",
        ["SubBase"]: array mixed[1]{"RRR"}
    },
    ["LRB"]  : dictionary{
        ["Index"]   : array[3]{"Left", "Right", "Base"}
        ["Type"]    : "SubBase",
        ["Centroid"]: <0,0,0>
    },
    ["RRR"]  : dictionary{
        ["Index"]   : array[3]{"Right1", "Right2", "Right"}
        ["Type"]    : "SubBase",
        ["Centroid"]: <0,0,0>
    }
  };
  #declare S = FABRIK(F);
  //StickMan(S) 

  #declare T = dictionary {
    ["Index"] : array[3]{"Left", "Right1", "Right2"},
    ["Right1"]: array mixed[1]{<5,2.5, 0>},
    ["Right2"]: array mixed[1]{<1,3.3, 0>},
    ["Left"]  : array mixed[1]{<-5, 3, 0>},
  };
  SetTarget(S, T)
  //StickMan(S) 
  //SplineMan(S) 
  PinMan(S)
#end

#end //test scene